#include "api/df_connection.h"
#include "datafox_library_iv.h"

DF_API_EXTERN_DECORATE HANDLE DF_API_CALL DFCGetComPort( int connectionId )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return INVALID_HANDLE_VALUE;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, connectionId );

    HANDLE hRes = channelBridge->getComPort();

    channelBridge->logEnd( "\t(%i, %i)", hRes, channelBridge->getLastErrorNumber( 0 ) );

    return hRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetComPort( int connectionId, HANDLE handle, int isSocket, int value, int timeout )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, connectionId );

    int iRes = (channelBridge->setComPort( handle, isSocket, value, timeout ) == IV_ERROR_SUCCESS);

    channelBridge->logEnd( "\t%i\t%i\t%i\t%i\t(%i, %i)", handle, isSocket, value, timeout, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComOpenSerial( int connectionId, const char *commPort, int baudRate, int timeout )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%s\t%i\t%i", __func__, connectionId,
        (commPort == nullptr) ? "" : commPort, baudRate, timeout );

    int iRes = FALSE;

    // Schnittstelle geschlossen?
    if ( channelBridge->isOpen() == false )
    {
        if ( channelBridge->openAEIII( 1, commPort, baudRate, timeout ) == IV_ERROR_SUCCESS )
        {
            iRes = TRUE;
        }
    }
    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComOpenSocket( int connectionId, const char *hostAddress, int portNumber, int timeout )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%s\t%i\t%i", __func__, connectionId,
        (hostAddress == nullptr) ? "" : hostAddress, portNumber, timeout );

    // zulaessige Bereiche pruefen
    if ( hostAddress == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    int iRes = FALSE;

    // Schnittstelle geschlossen?
    if ( channelBridge->isOpen() == false )
    {
        if ( channelBridge->openAEIII( 3, hostAddress, portNumber, timeout ) == IV_ERROR_SUCCESS )
        {
            iRes = TRUE;
        }
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComOpenIV( int connectionId, int deviceAddress, int connectionType, const char *connectionString, int connectionValue, int comunicationReadTimeout )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    if ( channelBridge->tryEnterCriticalSection() == false )
    {
        channelBridge->logBegin( "%s.ERROR\t\"Object in use. Owner-Id: %u\"", __func__, channelBridge->threadIdOpen_ );
        channelBridge->logEnd( "" );
        channelBridge->enterCriticalSection();
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i\t%i\t%s\t%i\t%i", __func__, connectionId, deviceAddress, connectionType,
        (connectionString == nullptr) ? "" : connectionString, connectionValue, comunicationReadTimeout );

    // zulaessige Bereiche pruefen
    if ( (connectionType != 1 && connectionType != 3 && connectionType != 4) 
        || connectionString == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( 0 ) );
        channelBridge->leaveCriticalSection();
        return 0;
    }


    int iRes = 0;

    // Schnittstelle geoschlossen?
    if ( channelBridge->isOpen() == false )
    {
        //2009_09_02: Busnummer fix auf 32
        //2012_05_21: Busnummer fix auf 0 (wegen Mehrfachdocking TimeboyIV)
        if ( channelBridge->open( 0, connectionType, connectionString, connectionValue, comunicationReadTimeout ) == IV_ERROR_SUCCESS )
        {
            iRes = TRUE;
            // Die Thread-Id halten und beim Schließen prüfen,
            // dass auch der gleiche Thread wieder die Verbindung beendet.
            channelBridge->threadIdOpen_ = std::this_thread::get_id();
        }
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );
    channelBridge->leaveCriticalSection();

    return iRes;
}

DF_API_EXTERN_DECORATE void DF_API_CALL DFCComClose( int connectionId )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, connectionId );

    if ( channelBridge->threadIdOpen_ != std::thread::id() && channelBridge->threadIdOpen_ != std::this_thread::get_id() )
    {
        if ( channelBridge->m_check_open_and_close_by_same_thread )
        {
            channelBridge->logBegin( "%s.ERROR\t\"Not object owner. Owner-Id: %u\"", __func__, channelBridge->threadIdOpen_ );
            channelBridge->logEnd( "" );
        }
        else
        {
            channelBridge->logBegin( "%s.WARNING\t\"Not object owner. Owner-Id: %u\"", __func__, channelBridge->threadIdOpen_ );
            channelBridge->logEnd( "" );

            channelBridge->close();
            channelBridge->threadIdOpen_ = std::thread::id();
        }
    }
    else
    {
        channelBridge->close();
        channelBridge->threadIdOpen_ = std::thread::id();
    }

    channelBridge->logEnd( "" );
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCWrite( int connectionId, const char *buffer, int toWrite, int *written, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, connectionId );

    // zulaessige Bereiche pruefen
    if ( buffer == nullptr || written == nullptr || errorNumber == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( 0 );
        }

        channelBridge->logEnd( "\t\t%i\t\t%i\t(%i, %i)", toWrite, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    int iRes = 0;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( 0 );
        
        channelBridge->logEnd( "\t%i\t\t%i\t(%i, %i)", toWrite, *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    if ( channelBridge->write( (const uint8_t *)buffer, toWrite, (std::size_t &)*written) == IV_ERROR_SUCCESS )
    {
        iRes = 1;
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t(%i, %i)", (strlen(buffer) <= toWrite) ? buffer : "",
        toWrite, *written, *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCRead( int connectionId, char *buffer, int toRead, int *read, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, connectionId );

    // zulaessige Bereiche pruefen
    if ( buffer == nullptr || read == nullptr || errorNumber == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( 0 );
        }

        channelBridge->logEnd( "\t\t%i\t\t%i\t(%i, %i)", toRead, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( 0 );
        
        channelBridge->logEnd( "\t%i\t\t%i\t(%i, %i)", toRead, *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    int iRes = 0;

    if ( channelBridge->read( (uint8_t *)buffer, toRead, (std::size_t &)read) == IV_ERROR_SUCCESS )
    {
        iRes = 1;
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t(%i, %i)", (strlen(buffer) <= toRead) ? buffer : "",
        toRead, *read, *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCIsChannelOpen( int connectionId )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr ) 
    {
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    int iRes = -1;
    channelBridge->logBegin( "%s\t%i", __func__, connectionId );

    if ( connectionId > 0 && connectionId < MAX_CHANNEL )
    {
        std::lock_guard<std::recursive_mutex> guard( theApp._mutexMapChannelBridge );
        if ( theApp._mapChannelBridge.count( connectionId ) )
        {
            if ( theApp._mapChannelBridge[connectionId]->isOpen() )
            {
                iRes = 1;
            }
            else 
            {
                iRes = 0;
            }
        }
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComUseHandle( int connectionId, HANDLE commHandle )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i", __func__, connectionId, commHandle );

    int iRes = (channelBridge->comUseHandle( commHandle ) == IV_ERROR_SUCCESS);

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComReleaseHandle( int connectionId )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, connectionId );

    int iRes = (channelBridge->comReleaseHandle() == IV_ERROR_SUCCESS);

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComInit( int connectionId )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i", __func__, connectionId );

    char szComStr[14];

#if defined(_MSC_VER) || defined(__MINGW32__)
    if ( connectionId > 9 )
    {
        sprintf( szComStr, "\\\\.\\COM%i", connectionId );
    }
    else
    {
        sprintf( szComStr, "COM%i", connectionId );
    }

#else
    sprintf( szComStr, "/dev/ttyS%i", connectionId - 1 );
#endif

    int iRes = FALSE;
    // Schnittstelle geschlossen?
    if ( channelBridge->isOpen() == false )
    {
        if ( channelBridge->openAEIII( 1, szComStr, 38400, 800 ) == IV_ERROR_SUCCESS )
        {
            iRes = TRUE;
        }
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComOpen( int connectionId, const char *hostAddress )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%s", __func__, connectionId, (hostAddress == nullptr) ? "" : hostAddress );

    if ( hostAddress == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    int iRes = FALSE;
    // Schnittstelle geschlossen?
    if ( channelBridge->isOpen() == false )
    {
        if ( channelBridge->openAEIII( 3, hostAddress, 8000, 5000 ) == IV_ERROR_SUCCESS )
        {
            iRes = TRUE;
        }
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

